;;;-*-Mode: LISP; Package: CCL -*-
; make-ccl2-help-file.lisp  Copyright 1990, Apple Computer, All Rights Reserved.

(in-package :ccl)

;;;;;;;;;;;;;;;;;;;;;
;;
;; Change History
;;
;; 10/01/96 slh  read-fred-docs-file: don't skip the last entry in the file
;; 03/02/96 bill Don't (require 'ff), read-fred-docs-file calls %buffer-set-read-only
;;               to undo the effect of the CKID resource.
;; ------------- 3.0
;; 02/24/92 bill remove unused binding from read-fred-docs-file
;;               close-help-resource-file
;; 02/07/92 bill Close help file while creating a new one.
;; 02/06/91 bill "MCL Help Map.lisp" -> "library;"
;; 01/28/91 bill "MCL Help Map.lisp" -> "examples;"
;;
;;;;;;;;;;;;;;;;;;;;;

;;this code converts Fred files with the following format:
;;each definition takes up two lines, no more or less.
;;  the first line contains the symbol name, followed by
;;  a space and arglist (if appropriate), followed by
;;  a tab followed by the documantation type in square brackets.
;;  (the documentation type could be "Function" or "Window Function"
;;  or "Variable", etc.
;;
;;  the second line contains the documentation string.
;;
;;  there are no empty lines in the file.
;;

;; Output is done by lib;simple-db.lisp
;; Output file has symbol names as keys.
;; Each entry looks like:
;;
;;  symbol-name
;;  type
;;  arglist
;;  documentation
;;
;; Font changes in the output are denoted by @x, where x is one of:
;;
;;  p  plain text, code font
;;  b  bold, code font
;;  d  plain text, documentation font
;;  i  italic, documentation font
;;
;; The code font will be Geneva 10 (or 9, if 10 is not available)
;; The documentation font will be New York 10.

;; MAKE-FRED-DOCS-FILE creates the help file: "ccl;MCL Help"
;; MAKE-HELP-MAP creates the help map file: "ccl;MCL Help Map.fasl"

(defparameter *help-maker-dir* "ccl:help-maker;")
(defparameter *fred-docs-dir* (concatenate 'string *help-maker-dir* "fred format;"))
(defparameter *docs-file* "ccl:MCL Help")

(defun get-doc-string (key)
  (let ((entries (get-index-file-entries *docs-file* key 1 nil)))
    (and entries (car entries))))

(defun make-fred-docs-file ()
  ; We check for exported symbols in these two modules
  (require 'quickdraw)
  #-ppc-target
  (require 'ff)
  (let ((docs-alist (read-fred-docs)))
    (write-fred-docs docs-alist)))

(defvar *bad-doc-names* nil)

(defun read-fred-docs ()
  (setq *bad-doc-names* nil)
  (let ((files (directory (merge-pathnames "*.*" *fred-docs-dir*)))
        res)
    (dolist (file files)
      (print file)
      (setq res (nconc (read-fred-docs-file file) res)))
    (when *bad-doc-names*
      (setq *bad-doc-names* (sort *bad-doc-names* 'string-lessp))
      (format t "~&See *BAD-DOC-NAMES*~%"))
    res))

(defun write-fred-docs (docs-alist)
  (let ((*arglist-on-space* nil))
    (when (streamp *doc-string-stream*)
      (close *doc-string-stream*))
    (if (fboundp 'close-help-resource-file)
      (funcall 'close-help-resource-file))
    (delete-file *docs-file*)
    (make-index-file *docs-file* docs-alist)
    (set-mac-file-type *docs-file* :HELP)))

(defconstant *bold-mask* (cdr (assq :bold *style-alist*)))
(defconstant *italic-mask* (cdr (assq :italic *style-alist*)))

(defun doc-font-id (ff ms)
  (let ((face (%ilsr 8 (point-h ff)))
        (size (point-h ms)))
    (cond ((not (eql 0 (logand face *bold-mask*)))
           #\b)
          ((not (eql 0 (logand face *italic-mask*)))
           #\i)
          ((eql size 9)
           #\p)
          (t #\d))))

(defvar *docs-buf* (make-buffer))

(defun string-remove-font-changes (string)
  (let ((len (length string))
        (len-diff 0)
        (pos 0))
    (loop
      (setq pos (%str-member #\@ string pos (- len len-diff)))
      (unless pos
        (return))
      (let ((del-count (if (or (>= (1+ pos) len)
                               (eql #\@ (schar string (1+ pos))))
                         1
                         2)))
        (incf len-diff del-count)
        (do ((i (+ pos del-count) (+ i 1)))
            ((>= i len))
          (setf (schar string (- i del-count)) (schar string i)))
        (incf pos (- 2 del-count))))
    (if (eql len-diff 0)
      string
      (subseq string 0 (- len len-diff)))))

(defun read-fred-docs-file (file)
  (let ((buf *docs-buf*))
    (buffer-delete buf 0 t)
    (buffer-insert-file buf file)
    (%buffer-set-read-only buf nil)
    (insert-font-changes buf)
    (let ((starts (canonicalize-doc-buf buf))
          (res nil))
      (while starts ; NOT (cddr starts)
        (let ((nameptr (pop starts))
              (docptr (pop starts))
              (nextptr (car starts)))
          (if (and (eql #\@ (buffer-char buf nameptr))
                   (not (eql #\@ (buffer-char buf (1+ nameptr)))))
            (incf nameptr 2))
          (let* ((name (string-remove-font-changes 
                        (string-upcase (buffer-substring buf nameptr (1- docptr))))))
            (multiple-value-bind (sym type) (find-symbol name :cl-user)
              (declare (ignore sym))
              (unless (and type
                           (or (eq type :external) 
                               (eq type :inherited)))
                (push name *bad-doc-names*)))
            (push
             (cons name
                   (buffer-substring buf docptr nextptr))
             res))))
      (nreverse res))))

(defun insert-font-changes (buf)
  (let ((pos 0)
        (id nil)
        (mark (make-mark buf))
        ff ms)
    (set-mark mark 0)
    (loop
      (setq pos (buffer-forward-find-char mark #\@ pos))
      (unless pos (return))
      (multiple-value-setq (ff ms) (buffer-char-font-codes buf (1- pos)))
      (set-mark mark pos)
      (buffer-set-font-codes mark ff ms)
      (buffer-insert mark #\@)
      (incf pos))
    (setq pos 0)
    (set-mark mark 0)
    (loop
      (multiple-value-setq (ff ms) (buffer-char-font-codes buf pos))
      (let* ((new-id (doc-font-id ff ms)))
        (unless (eql new-id id)
          (setq id new-id)
          (set-mark mark pos)
          (buffer-set-font-codes mark ff ms)
          (buffer-insert mark #\@)
          (buffer-insert mark id)
          (incf pos 2)))
      (unless (setq pos (buffer-next-font-change buf pos))
        (return)))))

(defun cons-string (&rest chars)
  (make-array (length chars) :element-type 'character :initial-contents chars))

(defvar *tab&cr* (cons-string #\tab #\newline))
(defvar *tab&cr&sp* (cons-string #\tab #\newline #\space))

(defun canonicalize-doc-buf (buf)
  (let ((mark (make-mark buf))
        namepos docpos p c res)
    (set-mark mark 0)
    (loop
      (setq namepos (buffer-position mark))
      (when (>= namepos (buffer-size mark)) (return))
      (when (and (eql #\@ (buffer-char mark))
                 (not (eql #\@ (buffer-char mark (1+ namepos)))))
        (buffer-delete mark namepos (+ 2 namepos)))
      (setq p (buffer-forward-find-char mark *tab&cr&sp*))
      (unless p (return))
      (set-mark mark (decf p))
      (setq c (buffer-char mark))
      (buffer-delete mark p (1+ p))
      (when (and (> p 2) (eq #\@ (buffer-char buf (- p 2))))
        (if (eq #\@ (buffer-char mark))
          (buffer-delete mark (- p 2) p)
          (set-mark mark (decf p 2))))
      (buffer-insert mark #\return)
      (setq docpos (buffer-position mark))
      (if (eql c #\tab)
        (progn
          (buffer-insert mark #\space)
          (buffer-insert mark #\return)
          (setq p (buffer-forward-find-char mark #\newline))
          (unless p (return))
          (set-mark mark p))
        (progn
          (setq p (buffer-forward-find-char mark *tab&cr*))
          (unless p (return))
          (set-mark mark (decf p))
          (setq c (buffer-char mark))
          (buffer-delete mark p (1+ p))
          (buffer-insert mark #\return)
          (if (not (eql c #\tab))
            (multiple-value-bind (ff ms) (buffer-char-font-codes mark)
              (let* ((id (doc-font-id ff ms))
                     (d? (eql id #\d)))
                (unless d?
                  (buffer-set-font-spec mark '("New York" 10 :plain))
                  (buffer-insert mark "@d"))
                (buffer-insert mark "[]")
                (unless d?
                  (buffer-set-font-codes mark ff ms)
                  (buffer-insert mark #\@)
                  (buffer-insert mark id)))
              (buffer-insert mark #\return))
            (progn
              (setq p (buffer-forward-find-char mark #\return))
              (unless p (return))
              (set-mark mark p)))))
      (setq p (buffer-forward-find-char mark #\return))
      (unless p (return))
      (set-mark mark p)
      (push namepos res)
      (push docpos res))
    (nreverse res)))
        
; MAKE-HELP-MAP is now in "library;make-help-map.lisp"
(require 'make-help-map)

(defvar *undocumented-lisp-symbols* nil)
(defvar *undocumented-ccl-symbols* nil)

(defun find-undocumented-symbols ()
  (let ((lisp-syms nil)
        (ccl-syms nil))
    (do-external-symbols (s :cl)
      (unless (gethash s *fast-help*)
        (push s lisp-syms)))
    (do-external-symbols (s :ccl)
      (unless (or (gethash s *fast-help*)
                  (memq (char (symbol-name s) 0) '(#\_ #\$)))
        (push s ccl-syms)))
    (setq *undocumented-lisp-symbols* (sort lisp-syms 'string-lessp)
          *undocumented-ccl-symbols* (sort ccl-syms 'string-lessp))
    (append *undocumented-ccl-symbols* *undocumented-lisp-symbols*)))

    
